#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2020  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

from jinja2 import Template,Environment, FileSystemLoader
import json
import sys
import os
import time
import ctypes
import requests
import logging
import re
import datetime

from waptserver.config import wapt_root_dir

logger = logging.getLogger('glpiinventory.inventory')

def srn(serialnumber):
    if serialnumber >= 0:
        return hex(serialnumber)[2:-1].upper()
    else:
        return hex(int(ctypes.c_uint32(serialnumber).value))[2:-1].upper()

def get_cpu_infos(data):
    cpu = {}
    items_key = ['Core_Count', 'Thread_Count', 'ID', 'Max_Speed', 'Manufacturer', 'Serial_Number']
    items = ['core', 'thread', 'id', 'speed', 'manufacturer', 'serial']

    stepping = re.split('stepping', data['host_info']['cpu_identifier'], flags=re.IGNORECASE)
    if len(stepping) > 1:
        cpu['stepping'] = [v.strip() for v in stepping[1].split(' ') if v.isdigit()][0]
    else:
        cpu['stepping'] = ''

    for i in range(len(items)):
        if not items_key[i] in data['dmi']['Processor_Information'][0].keys():
            continue
        cpu[items[i]] = data['dmi']['Processor_Information'][0][items_key[i]]

    cpu['speed'] = cpu['speed'].replace(' MHz','')
    cpu['name'] = data['host_info']['cpu_name']
    cpu['model'] = data['host_info']['cpu_identifier']
    try:
        val = data['dmi']['Processor_Information'][0]['Signature']
        if 'Family' in val:
            cpu['familynumber'] = val.split('Family')[1].split(',')[0].strip()
            cpu['modelnumber'] = val.split('Model')[1].split(',')[0].strip()
    except:
        pass
    if 'intel' in cpu['manufacturer'].lower():
        cpu['manufacturer'] = 'Intel'

    return cpu

def get_ram_infos(data):
    ram = []
    nb = 0
    for entry in data['dmi']['Memory_Device'] :
        d = {}
        nb = nb + 1

        d["CAPACITY"] = entry.get("Size",'')
        if d["CAPACITY"]:
            if "mb" in d["CAPACITY"].lower():
                d["CAPACITY"] = d["CAPACITY"].split(' ')[0]
            elif "gb" in d["CAPACITY"].lower():
                d["CAPACITY"] = int(d["CAPACITY"].split(' ')[0]) * 1024
            elif "tb" in d["CAPACITY"].lower():
                d["CAPACITY"] = int(d["CAPACITY"].split(' ')[0]) * 1024 * 1024
            else:
                d["CAPACITY"] = d["CAPACITY"].split(' ')[0]

        d['CAPTION'] = entry.get("Locator",'')
        d['DESCRIPTION'] = entry.get("From_Factor",'')
        d['MANUFACTURER'] = entry.get("Manufacturer",'')
        d['MODEL'] = entry.get("Part_Number",'')
        d['NUMSLOTS'] = str(nb)
        d['SERIALNUMBER'] = entry.get("Serial_Number",'')
        d["SPEED"] = entry.get("Speed",'')
        d['TYPE'] = entry.get("Type",'')
        ram.append(d)
    return ram

class HostInventory():

    def __init__(self):
        self.data = {}
        self.date = time.strftime("%Y-%m-%d")+" "+time.strftime("%H:%M:%S")
        self.inventory_data = None

    def data_refactoring(self):
        if not self.data:
            logger.error('Inventory not created, no data to refactor')
            return -1

        newsoft = []
        for soft in self.data['installed_softwares']:
            try:
                soft['install_date']= datetime.datetime.strptime(soft['install_date'], '%Y-%m-%d %H:%M:%S').strftime('%d/%m/%Y') 
            except:
                pass
            newsoft.append(soft)

        self.data['installed_softwares'] = newsoft           

        if type(self.data['wmi']['Win32_LogicalDisk']) == dict:
            self.data['wmi']['Win32_LogicalDisk'] = [self.data['wmi']['Win32_LogicalDisk']]
        if type(self.data['wmi']['Win32_Volume']) == dict:
            self.data['wmi']['Win32_Volume'] = [self.data['wmi']['Win32_Volume']]
        if type(self.data['wmi']['Win32_NetworkAdapter']) == dict:
            self.data['wmi']['Win32_NetworkAdapter'] = [self.data['wmi']['Win32_NetworkAdapter']]
        if type(self.data['wmi']['Win32_Printer']) == dict:
            self.data['wmi']['Win32_Printer'] = [self.data['wmi']['Win32_Printer']]
        if type(self.data['dmi']['Processor_Information']) == dict:
            self.data['dmi']['Processor_Information'] = [self.data['dmi']['Processor_Information']]
        if type(self.data['dmi']['Memory_Device']) == dict:
            self.data['dmi']['Memory_Device'] = [self.data['dmi']['Memory_Device']]

        if self.data['host_info'].get('workgroup_name','') in ('AUTORITE NT','NT AUTHORITY'):
            self.data['host_info']['workgroup_name'] = self.data['host_info']['computer_name']

        self.data['host_metrics']['physical_memory'] = int( int(self.data['host_metrics'].get('physical_memory','0')) /1024/1024 )

        self.data['is_virtual'] = ''
        if self.data['wmi']['Win32_ComputerSystem']['Model']:
            model = self.data['wmi']['Win32_ComputerSystem']['Model']
            self.data['is_virtual'] = int ( (any(vm_type in model.lower() for vm_type in ('vmware','hvm','virtual')) == True ))

        if self.data['host_metrics'].get('last_logged_on_user'):
            if isinstance(self.data['host_metrics'].get('last_logged_on_user'),list):
                self.data['host_metrics']['last_logged_on_user'] = self.data['host_metrics']['last_logged_on_user'][0].split('\\')[-1]
            else:
                self.data['host_metrics']['last_logged_on_user'] = self.data['host_metrics']['last_logged_on_user'].split('\\')[-1]

        self.data['typedisk'] = ['Unknown','No Root Directory','Removable Disk','Local Disk','Network Drive','Compact Disc','RAM Disk']
        self.data['cpu'] = get_cpu_infos(self.data)  # TODO reconstruction of cpu informations from database to json
        self.data['ram'] = get_ram_infos(self.data)
        self.data['allip'] = '/'.join([addr for addr in self.data['host_info']['connected_ips'] if addr])


        if self.data['wmi']['Win32_OperatingSystem']['OSArchitecture']:
            if self.data['wmi']['Win32_OperatingSystem']['OSArchitecture']=='64 bits':
                self.data['wmi']['Win32_OperatingSystem']['OSArchitecture']='64-bit'

        if self.data['wmi']['Win32_OperatingSystem']['LastBootUpTime']:
            val = self.data['wmi']['Win32_OperatingSystem']['LastBootUpTime']
            self.data['wmi']['Win32_OperatingSystem']['LastBootUpTime'] =  '%s-%s-%s %s:%s:%s' % (val[0:4],val[4:6],val[6:8],val[8:10],val[10:12],val[12:14])

        if self.data['wmi']['Win32_OperatingSystem']['InstallDate']:
            val = self.data['wmi']['Win32_OperatingSystem']['InstallDate']
            self.data['wmi']['Win32_OperatingSystem']['InstallDate'] = '%s-%s-%s %s:%s:%s' % (val[0:4],val[4:6],val[6:8],val[8:10],val[10:12],val[12:14])

        if self.data['host_info']['windows_version']:

            corresp = {'10240':'1507',
                       '10586':'1511',
                       '14393':'1607',
                       '15063':'1703',
                       '16299':'1709',
                       '17134':'1803',
                       '17763':'1809',
                       '18362':'1903',
                       '18363':'1909',
                       '19041':'2004',
                       '19042':'20H2',
                       '19043':'21H1',
                       '19044':'21H2'
                       }
            try:

                if 'windows_version_releaseid' in self.data['host_info'] :
                    self.data['operatingsystem_version']=self.data['host_info']['windows_version_releaseid']

                if "windows_version_prettyname" in self.data['host_info']:
                    self.data['operatingsystem_version']=self.data['host_info']['windows_version_prettyname']

                if not 'operatingsystem_version' in self.data:
                    self.data['operatingsystem_version']=corresp[ self.data['host_info']['windows_version'].split('.')[-1] ]

            except:
                pass
        if self.data['wmi']['Win32_OperatingSystem']['CSDVersion'] == None:
            self.data['wmi']['Win32_OperatingSystem']['CSDVersion'] = ''
        if 'windows' in self.data['host_info']['os_name'].lower():
            self.data['host_info']['os_shortname']='Windows'

        if 'Portable_Battery' in self.data['dmi']:
            orig_data = self.data['dmi']['Portable_Battery']
            temp_list = list()
            if isinstance(orig_data,dict):
                orig_data = [orig_data]
            for battery in orig_data:
                battery['Design_Capacity'] = battery['Design_Capacity'].replace('mWh','').strip()
                try:
                    val = battery['SBDS_Manufacture_Date']
                except:
                    val = battery['Manufacture_Date']

                battery['SBDS_Manufacture_Date'] = '%s/%s/%s' % (val[8:10],val[5:7],val[0:4])

                try:
                    val = battery['SBDS_Serial_Number']
                except:
                    val = battery['Serial_Number']
                try:
                    battery['SBDS_Serial_Number'] = int(val, 16)
                except:
                    pass
                battery['Design_Voltage'] = battery['Design_Voltage'].replace('mV','').strip()
                temp_list.append(battery)
            self.data['dmi']['Portable_Battery'] = temp_list

        self.data['host_metrics']['last_bootup_time']=self.data['host_metrics']['last_bootup_time'].replace('T',' ')

    #        for env_var in data['host_info']['environ']:
        if self.data['wmi']['Win32_OperatingSystem']['Caption']:
            self.data['host_info']['os_name'] = self.data['wmi']['Win32_OperatingSystem']['Caption']

        dict_mac_addr = {}
        for network in self.data['wmi']['Win32_NetworkAdapter']:
            if network.get('MACAddress'):
                dict_mac_addr[network['MACAddress']] = {"ip":"","netmask":""}
                for interface in self.data['host_info']['networking']:
                        if interface['mac'].lower() == network['MACAddress'].lower():
                            if interface.get('addr') and type(interface['addr']) == list:
                                dict_mac_addr[network['MACAddress']] = {"addr":interface['addr'][0]['addr'],"netmask":interface['addr'][0]['netmask']}
                                break

        self.data['dict_mac_addr'] = dict_mac_addr

        if self.data['dmi']['System_Information']['Manufacturer'] == 'LENOVO':
            self.data['smodel'] = self.data['dmi']['System_Information']['Version']
        else:
            self.data['smodel'] = self.data['dmi']['System_Information']['Product_Name']

        return 0

    def create_inventory(self, data):
        self.data = {}
        self.data = data
        if self.data_refactoring() == -1:
            return
        template_dir = os.path.join(wapt_root_dir, 'waptserver', 'templates')
        TEMPLATE_ENVIRONMENT = Environment(
                            autoescape=True,
                            loader=FileSystemLoader(template_dir),
                            trim_blocks=False)
        self.inventory_data = TEMPLATE_ENVIRONMENT.get_template('glpi-inventory.j2').render(data=self.data,date=self.date,srn=srn)

    def write_inventory(self, file='inventory_debug.xml'):
        if not self.inventory_data:
            logger.error('Inventory data not created, please create_inventory()')
            return
        with open(file,'w', encoding='utf-8') as outfile:
            outfile.write(self.inventory_data)

    def upload_to_glpi(self, username, password, endpoint, headers=None, quiet=False):
        if not self.inventory_data:
            logger.error('Inventory data not created, please create_inventory()')
            return
        if not headers:
            headers = {'charset':'utf-8','Pragma' :'no-cache',"content-type":"Application/xml"}

        creds = (username.encode(encoding="UTF-8"), password.encode(encoding="UTF-8"))
        r = requests.post(endpoint, data=self.inventory_data.encode('utf-8'),  headers=headers, auth=creds)
        if not quiet:
            logger.info(r.content)
        return r


