#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2020  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##
import os
import sys
import re
import datetime
import time
import json
import base64

from waptutils import Version, datetime2isodate

from peewee import *

# flask
from flask import request, Flask, Response, send_from_directory, send_file, session
from flask import redirect, url_for, abort, render_template, flash, safe_join, Blueprint
from werkzeug.utils import secure_filename

from waptserver.app import app

from waptserver.model import Packages, load_db_config, StoreDownload, StoreMember, StoreUsage, wapt_db, WaptDB
from waptpackage import PackageEntry,PackageVersion
from waptserver.decorators import requires_auth, allow_local, require_wapt_db, requires_enterprise

from wtforms import Form, TextField, TextAreaField, validators, StringField, SubmitField, BooleanField, PasswordField

import string
import random
import json
import logging

# i18n
from flask_babel import Babel
try:
    from flask_babel import gettext
except ImportError:
    gettext = (lambda s: s)

_ = gettext


logger = logging.getLogger('store')

store = Blueprint('store', __name__, template_folder='templates')


babel = Babel(app)

BASE_DIR = os.path.join(app.conf['wapt_folder'])

try:
    from flask_babel import gettext
except ImportError:
    gettext = (lambda s: s)

_ = gettext


class RegistrationFrom(Form):
    surname = StringField('Surname', validators=[validators.required(), validators.Length(min=2, max=40)])
    firstname = StringField('Firstname', validators=[validators.required(), validators.Length(min=2, max=40)])
    organization = StringField('Organization', validators=[validators.required(), validators.Length(min=2, max=50)])
    email = StringField('Email', validators=[validators.required(), validators.Length(min=6, max=40)])
    password = PasswordField('New Password', [
        validators.required(),
        validators.EqualTo('confirm_password', message='Passwords must match'),
        validators.Length(min=7, max=30)
    ])
    confirm_password = PasswordField('Repeat Password', validators=[validators.required()])
    accept_gdpr = BooleanField('RDPR')
    accept_tos = BooleanField('TOS', validators=[validators.required()])


@babel.localeselector
def get_locale():
    browser_lang = request.accept_languages.best_match(['en', 'fr'])
    user_lang = session.get('lang', browser_lang)
    return user_lang


def format_isodate(isodate):
    """Pretty format iso date like : 2014-01-21T17:36:15.652000
        >>> format_isodate('2014-01-21T17:36:15.652000')
        '21/01/2014 17:36:15'
    """
    return datetime2isodate(isodate).replace('T', ' ')[0:16]


def count_packages():
    return Packages.select(fn.COUNT(Packages.package)).where(Packages.section == 'base').scalar() or 0


def get_locales():
    return [l[0] for l in Packages.select(Packages.locale.distinct()).where(Packages.locale != '', Packages.locale != 'all', Packages.locale != 'fr', Packages.locale != 'en').order_by(Packages.locale).tuples()]


def get_os(package):
    return list(Packages.select(Packages.target_os).where((Packages.package == package) & (Packages.target_os != None)).group_by(Packages.target_os).dicts())


def count_download(package):
    """Returns total download count for a package

    Args:
        package (str): package name

    Returns:
        int : download count
    """
    return StoreDownload.select(fn.SUM(StoreDownload.download_number)).where(StoreDownload.package_name == package).scalar() or 0


def netname(ip):
    whois = os.popen('whois %s | grep netname' % ip).read()
    name = (whois and whois.split(':')[1].strip()) or 'unknown'
    return name


def ptr(ip):
    ptr = os.popen('dig +short -t ptr -x %s' % ip).read().splitlines()
    if ptr:
        ptr = ptr[0].strip()
    else:
        ptr = 'unknown'
    return ptr


def most_popular():
    SQL_most_popular = '''\
            SELECT packages.package, version, architecture, maturity, locale, signature_date, description, categories, target_os, icon_sha256um FROM packages
            LEFT JOIN StoreDownload
            ON packages.package = StoreDownload.package_name
            WHERE section='base' AND packages.id IN
                (SELECT DISTINCT ON(package) id FROM packages
                WHERE version IN
                    (SELECT DISTINCT ON(package) version FROM packages
                    GROUP BY package,version ORDER BY package, MAX(string_to_array(split_part(version,'-',1), '.')::Bigint[]) DESC)
            GROUP BY package, id)
            GROUP BY package, version, architecture, maturity, locale, packages.signature_date, description, categories, target_os, icon_sha256sum
            ORDER BY SUM(StoreDownload.download_number) DESC LIMIT 4\
        '''
    return wapt_db.execute_sql(SQL_most_popular)


@store.route('/store')
@store.route('/store/')
@require_wapt_db
@requires_enterprise
def packages():
    """
    Homepage's store

    Args:
        sort(str): sort result by popular, name or date. Default value it's popular
        search(str): keyname to search a package. Default value is None
        arch(str): architecture type. Default value is 'x64, x86, all, "" '
        locale(str): language package. Default value is empty
        categorie(str): categorie package. Default value is empty

    Returns:
        str: html page

    """
    sortkey = request.args.get('sort', 'popular')
    search = request.args.get('search', None)
    archkey = request.args.getlist('arch')
    localekey = request.args.getlist('locale')
    categorie = request.args.getlist('categorie', None)
    targetoskey = request.args.getlist('targetos', None)

    session.pop('search', None)
    session.pop('arch', None)
    session.pop('locale', None)
    session.pop('categorie', None)
    session.pop('targetos', None)

    # Check all args. If agrs is empty or with no authorized string, args are cleaned and set 'all'.
    if set(archkey).issubset(set(['x64', 'x86', 'all'])):
        arch_list = archkey
        if len(archkey) == 2 and archkey[0] == 'all':
            archkey = "('%s','all','')" % archkey[1]
        elif len(archkey) == 1 and archkey[0] != 'all':
            archkey = "('%s','all','')" % archkey[0]
        else:
            archkey = "('x64','x86','all','')"
    else:
        archkey = "('x64','x86','all','')"
        arch_list = ''

    if localekey:
        if set(localekey).issubset(['fr', 'en', 'de', 'es', 'it', 'all', '']):
            locales_list = localekey
            loc_query = "','".join(localekey)
            localekey = "AND packages.locale IN ('', 'all', '%s')" % loc_query
        else:
            localekey = ''
            locales_list = ''
    else:
        localekey = ''
        locales_list = ''

    if search is None:
        search = ''

    if re.match('^[a-zA-Z0-9.-]+$', search):
        searchkey = '%s' % search
    else:
        search = ''
        searchkey = '%s' % search

    if set(categorie).issubset(set(['Internet', 'Utilities', 'Messaging', 'Security', 'Systemandnetwork', 'Media', 'Development', 'Office', 'Drivers'])):
        if len(categorie) > 0:
            categoriekey = 'AND ('
            categories_list = []
            if categorie[0] == 'Systemandnetwork':
                categorie[0] = "System and network"
            categoriekey = categoriekey + "'%s' = ANY (packages.categories)" % categorie[0]
            categories_list.append(categorie[0])
            if len(categoriekey) > 1:
                for cat in categorie[1:]:
                    if cat == 'Systemandnetwork':
                        cat = "System and network"
                    categoriekey = categoriekey + " OR '%s' = ANY (packages.categories)" % cat
                    categories_list.append(cat)
            categoriekey = categoriekey + ')'
        else:
            categoriekey = ''
            categories_list = ''
    else:
        categoriekey = ''
        categories_list = ''

    if targetoskey and set(targetoskey).issubset(set(['windows', 'linux', 'macos'])):
        for n, os in enumerate(targetoskey):
            if "macos" in os:
                targetoskey[n] = "darwin"

        targetoslist = targetoskey
        targetos_list = targetoskey
        os_query = "','".join(targetoskey)
        targetos_list = "AND packages.target_os IN ('', 'all', '%s')" % os_query
    else:
        targetoslist = ''
        targetos_list = ''

    # SQL query sort by alpha
    if sortkey == 'alphaasc':
        SQL = '''\
            SELECT DISTINCT ON(package) packages.package, version, architecture, maturity, locale, packages.signature_date, description, categories, target_os, icon_sha256sum FROM packages
            WHERE section='base' AND packages.package ILIKE '%%{package_search}%%' AND packages.architecture IN {arch_search} {locale_search} {categories_search} {targetos_search}
            GROUP BY packages.package, version, description, packages.signature_date, architecture, maturity, locale, categories, target_os, icon_sha256sum
            ORDER BY packages.package, MAX(string_to_array(split_part(version,'-',1), '.')::Bigint[])  \
        '''.format(package_search=searchkey, arch_search=archkey, locale_search=localekey, categories_search=categoriekey, targetos_search=targetos_list)
        packages = wapt_db.execute_sql(SQL)

    elif sortkey == 'alphadesc':
        SQL = '''\
            SELECT DISTINCT ON(package) packages.package, version, architecture, maturity, locale, packages.signature_date, description, categories, target_os, icon_sha256sum FROM packages
            WHERE section='base' AND packages.package ILIKE '%%{package_search}%%' AND packages.architecture IN {arch_search} {locale_search} {categories_search} {targetos_search}
            GROUP BY packages.package, version, description, packages.signature_date, architecture, maturity, locale, categories, target_os, icon_sha256sum
            ORDER BY packages.package DESC, MAX(string_to_array(split_part(version,'-',1), '.')::Bigint[])   \
        '''.format(package_search=searchkey, arch_search=archkey, locale_search=localekey, categories_search=categoriekey, targetos_search=targetos_list)
        packages = wapt_db.execute_sql(SQL)

    # SQL query sort by date
    elif sortkey == 'date':
        SQL = '''\
            SELECT DISTINCT ON(signature_date) packages.package, version, architecture, maturity, locale, packages.signature_date, description, categories, target_os, icon_sha256sum FROM packages
            WHERE section='base' AND packages.package ILIKE '%%{package_search}%%' AND packages.architecture IN {arch_search} {locale_search} {categories_search}  {targetos_search} AND packages.id IN
                (SELECT DISTINCT ON(package) id FROM packages
                    WHERE version IN
                    (SELECT DISTINCT ON(package) version FROM packages WHERE packages.architecture IN {arch_search} {locale_search} {categories_search}
                    GROUP BY package,version ORDER BY package, MAX(string_to_array(split_part(version,'-',1), '.')::Bigint[]) DESC) AND packages.architecture IN {arch_search} {locale_search} {categories_search}
                GROUP BY package, id)
            GROUP BY package, version, architecture, maturity, locale, signature_date, description, categories, target_os, icon_sha256sum
            ORDER BY signature_date DESC \
        '''.format(package_search=searchkey, arch_search=archkey, locale_search=localekey, categories_search=categoriekey, targetos_search=targetos_list)
        packages = wapt_db.execute_sql(SQL)

    # Default case, sort by popularity
    else:
        SQL = '''\
            SELECT packages.package, version, architecture, maturity, locale, signature_date, description, categories, target_os, icon_sha256sum FROM packages
            LEFT JOIN StoreDownload
            ON packages.package = StoreDownload.package_name
            WHERE section='base' AND packages.package ILIKE '%%{package_search}%%' AND packages.architecture IN {arch_search} {locale_search} {categories_search} {targetos_search} AND packages.id IN
                (SELECT DISTINCT ON(package) id FROM packages
                WHERE version IN
                    (SELECT DISTINCT ON(package) version FROM packages WHERE packages.architecture IN {arch_search} {locale_search} {categories_search} {targetos_search}
                    GROUP BY package,version ORDER BY package, MAX(string_to_array(split_part(version,'-',1), '.')::Bigint[]) DESC) AND packages.architecture IN {arch_search} {locale_search} {categories_search} {targetos_search}
            GROUP BY package, id)
            GROUP BY package, version, architecture, maturity, locale, packages.signature_date, description, categories, target_os, icon_sha256sum
            ORDER BY SUM(StoreDownload.download_number) DESC \
        '''.format(package_search=searchkey, arch_search=archkey, locale_search=localekey, categories_search=categoriekey, targetos_search=targetos_list)
        packages = wapt_db.execute_sql(SQL)

    SQL2 = '''\
        SELECT count(*) FROM packages
        WHERE section='base' AND packages.package ILIKE '%%{package_search}%%' AND packages.architecture IN {arch_search} {locale_search} {categories_search} {categories_search} {targetos_search}

    '''.format(package_search=searchkey, arch_search=archkey, locale_search=localekey, categories_search=categoriekey, targetos_search=targetos_list)
    nbre_packages = wapt_db.execute_sql(SQL2)

    """
    Print all queries in variables to transfer them in HTML template

    """
    try:
        packages_list = list(packages.dicts())
    except:
        packages_list = list(packages)

    nbre_packages = nbre_packages.fetchone()[0]

    if request.args.getlist('locale'):
        session['locale'] = request.args.getlist('locale')

    if archkey:
        session['arch'] = archkey

    if search != '':
        session['search'] = request.args.getlist('search')

    """
    Print result

    """
    if len(packages_list) >= 1:
        return render_template('home.html', packages=packages_list, format_isodate=format_isodate, count_download=count_download, count_packages=count_packages,
                               get_locales=get_locales, categories=categories_list, locales=locales_list, architectures=arch_list, nbre_packages=nbre_packages, get_os=get_os, targetos=targetoslist,
                               args_list=request.args)

    elif len(packages_list) == 0:
        error_message = "no_result"
        return render_template('home.html', packages=packages_list, format_isodate=format_isodate, count_download=count_download, count_packages=count_packages,
                               get_locales=get_locales, categories=categories_list, locales=locales_list, architectures=arch_list, nbre_packages=nbre_packages, get_os=get_os, targetos=targetoslist,
                               error_message=error_message, most_popular=most_popular, args_list=request.args)
    else:
        error_message = "not_found"
        return render_template('error.html', error_message=error_message, count_packages=count_packages, get_locales=get_locales, count_download=count_download, most_popular=most_popular, get_os=get_os,)


@store.route('/store/search', methods=['GET', 'POST'])
@requires_enterprise
@require_wapt_db
def search():
    if request.method == 'POST':
        search = request.form['search']
    else:
        search = request.args.get('search', None)

    if re.match('^[a-zA-Z0-9.-]+$', search):
        searchkey = '%%%s%%' % search
    else:
        search = ''
        searchkey = '%%%s%%' % search

    query = Packages.select(Packages.package).where((Packages.package ** searchkey) & (Packages.section == "base")).group_by(Packages.package).order_by(Packages.package).limit(15)
    result = list(query.dicts())

    return Response(response=json.dumps(result), status=200, mimetype='application/json')


@store.route('/store/<package>')
@requires_enterprise
@require_wapt_db
def list_package(package):
    if re.match('^[a-zA-Z0-9.-]+$',package):
        software = list(Packages.select().where(Packages.package == package).order_by(Packages.version.desc(),Packages.locale.asc()).dicts())
        target_os_list = list(Packages.select(Packages.target_os).where(Packages.package == package).group_by(Packages.target_os,Packages.package).dicts())

        if software:
            args = request.args

            for p in software:
                pe = PackageEntry(**p)
                if not pe.filename:
                    pe.filename = pe.make_package_filename()

            new_packages_list = []
            old_packages_list = []

            for os in target_os_list:
                last_version = '0-0'
                for soft in software:
                    if str(soft['target_os']).lower() == str(os['target_os']).lower():
                        soft_version = soft['version']

                        if Version(soft_version.split('-')[0]) >= Version(last_version.split('-')[0]):
                            last_version = soft_version
                            new_packages_list.append(soft)

                        if Version(soft_version) != Version(last_version):
                            old_packages_list.append(soft)

            return render_template('package_version.html', new_packages_list=new_packages_list,old_packages_list=old_packages_list,package=package, last_version=last_version,
                get_locales=get_locales ,format_isodate=format_isodate, count_download=count_download)

        else:
            error_message = "not_found"
            return render_template('error.html', error_message=error_message, count_packages=count_packages, get_locales=get_locales, most_popular=most_popular, count_download=count_download,
                                   get_os=get_os)
    else:
        error_message = "invalid_package"
        return render_template('error.html', error_message=error_message, count_packages=count_packages, get_locales=get_locales, most_popular=most_popular, count_download=count_download,
                               get_os=get_os)


@store.route('/store/details-<filename>')
@requires_enterprise
@require_wapt_db
def package_details(filename):
    package_path = safe_join(BASE_DIR, filename)
    if os.path.isfile(package_path):
        package = PackageEntry(waptfile=package_path)
        package.previous = None
        with package.as_zipfile() as zipf:
            filenames = [f.filename for f in zipf.filelist]
            taskid_luti = None
            folder_in_git_repo = None
            for filen in ['setup.py', 'WAPT/changelog.txt', 'WAPT/control', 'WAPT/manifest.sha256', 'luti.json']:
                if filen in filenames:
                    try:
                        package[filen] = zipf.open(filen, 'r').read().decode('utf8')
                    except:
                        logger.error('Encoding Error : ' + filen)
                        package[filen] = zipf.open(filen, 'r').read().decode('iso-8859-1')

                    if filen == 'luti.json':
                       # try:
                        lutijson = zipf.open('luti.json','r').read().decode('utf8')

                        taskid_luti = json.loads(lutijson)['taskid_luti']
                        folder_in_git_repo = base64.b64encode(json.loads(lutijson)['folder_in_git_repo'].encode('utf-8')).decode('utf-8')
##                        except:
##                            taskid_luti = "no_taskid_luti"


        return render_template('package_detail.html', package=package, package_name=filename, count_packages=count_packages,
                               get_locales=get_locales, count_download=count_download, taskid_luti=taskid_luti, folder_in_git_repo=folder_in_git_repo)
    else:
        error_message = "not_found"
        return render_template('error.html', error_message=error_message, count_packages=count_packages, get_locales=get_locales, most_popular=most_popular, count_download=count_download,
                               get_os=get_os)


@store.route('/store/package_details')
@requires_enterprise
@require_wapt_db
def package_details_old_link():
    filename = request.args.get('package', '')
    package_path = safe_join(BASE_DIR, filename)
    if os.path.isfile(package_path):
        return redirect('/store/details-%s' % filename, code=301)
    elif filename.rsplit('.')[0]:
        return redirect('/store/%s' % filename, code=301)
    else:
        error_message = "not_found"
        return render_template('error.html', error_message=error_message, count_packages=count_packages, get_locales=get_locales)


@store.route('/store/download/<filename>', methods=['GET'])
@requires_enterprise
@require_wapt_db
def download(filename):

    wapt_packages_local = os.path.join(BASE_DIR)

    package_name = filename
    filename_exception = ["Packages", "waptserversetup.exe", "waptsetup.exe", "waptstarter.exe", "waptdeploy.exe", "waptconsole.exe", "tranquilit.crt"]
    package_path = "%s%s" % (wapt_packages_local, package_name)
    package_dl = '/wapt/' + package_name
    _, package_ext = os.path.splitext(package_name)

    if package_name in filename_exception:
        pass

    elif package_ext == ".deb":
        package_path = '/var/www/wapt/debian/' + package_name

    elif package_name == 'Packages.gz':
        package_path = '/var/www/wapt/debian/' + package_name

    elif package_ext != '.wapt':
        logger.error('404 - No valid extention !')
        invalid_request = True
        error_message = "invalid_package"
        # exit()

    if not os.path.isfile(package_path):
        logger.error('404 - File not found ! ')
        invalid_request = True
        error_message = "not_found"
        # exit()

    if invalid_request:
        return render_template('error.html', error_message=error_message, count_packages=count_packages, get_locales=get_locales)
    else:

        try:
            software_version = package_name.split('_')[1].split('-')[0]
            software = package_name.split('_')[0]
        except:
            software = package_name.split('.')[0]

        if StoreDownload.select(StoreDownload.package).where(StoreDownload.package == software):
            StoreDownload.update(
                download=StoreDownload.download+1,
                last_download=datetime.datetime.now()
            ).where(
                StoreDownload.package == software
            ).execute()
        else:
            s = StoreDownload()
            s.package = software
            s.download = '1'
            s.last_download = datetime.datetime.now()
            s.save()

    return send_file(os.path.join(wapt_packages_local, package_name), as_attachment=True)


@store.route('/store/register', methods=['GET', 'POST'])
@requires_enterprise
@require_wapt_db
def register():
    form = RegistrationFrom(request.form)

    logger.info(form.errors)

    logger.info(request.method)

    if request.method == 'POST':
        logger.info(request.form)

        if request.form.get('surname', None):
            surname = request.form['surname']
        else:
            surname = ''
        firstname = request.form['firstname']
        organization = request.form['organization']
        email = request.form['email']
        password = request.form['password']
        confirm_password = request.form['confirm_password']

        if request.form.get('accept_gdpr'):
            accept_gdpr = True
        else:
            accept_gdpr = False

        accept_tos = True

        if form.validate():
            flash('Thanks for registration %s' % email)
            re_mem = StoreMember()
            re_mem.surname = surname
            re_mem.firstname = firstname
            re_mem.organization = organization
            re_mem.email = email
            re_mem.password = password
            re_mem.gdpr = accept_gdpr
            re_mem.tos = accept_tos
            re_mem.token = token_generator()
            re_mem.validat_account = False
            re_mem.save()

        else:
            flash('Errors are detected')

    return render_template('register.html', form=form, count_packages=count_packages, get_locales=get_locales)

@store.route('/store/packages_list')
@requires_enterprise
@require_wapt_db
def packages_list():
    SQL = """ SELECT DISTINCT ON(package) packages.package, version  FROM packages
              WHERE section='base' AND id IN
                  (SELECT DISTINCT ON(package) id FROM packages
                   WHERE version IN
                       (SELECT DISTINCT ON(package) version FROM packages
                        GROUP BY package,version ORDER BY package, MAX(string_to_array(split_part(version,'-',1), '.')::Bigint[]) DESC)
                   GROUP BY package,version,id ORDER BY package, MAX(string_to_array(split_part(version,'-',1), '.')::Bigint[]) DESC)
             GROUP BY packages.package, id
             ORDER BY packages.package, MAX(string_to_array(split_part(version,'-',1), '.')::Bigint[]) """

    res = wapt_db.execute_sql(SQL)

    try:
        result = list(res.dicts())
    except:
        result = list(res)


    return Response(response=json.dumps(result),
                      status = 200,
                      mimetype = "application/json")
